//
//  MINTLDocumentDefaults.swift
//  DocumentDefaultsSample
//
//  Created by narita on 2019/05/30.
//  Copyright © 2019 narita. All rights reserved.
//

import Cocoa

// 用語
// key    :
// value  :
// recode : key/valueのペア実態は、NSManagedObject
// entity : coreDataのTable名

extension String
{
    public static let MINTLDocumentDefaultsKeyNameString: String   = "key"
    public static let MINTLDocumentDefaultsValueNameString: String = "value"
}

class MINTLDocumentDefaults: NSObject
{
    // MARK: - class methods
    
    // MARK: - init methods
    public init(document: NSPersistentDocument, entityName: String)
    {
        super.init()
        
        self.document = document
        self.entityName = entityName
        
        // UserDefaultの変更を捉える
        NotificationCenter.default.addObserver(forName: UserDefaults.didChangeNotification,
                                               object: UserDefaults.standard,
                                               queue: nil)
        {
            _ in
            
            self.passThroughKeySet.forEach()
            {
                self.willChangeValue(forKey: $0)
                self.didChangeValue(forKey: $0)
            }
        }
        
        // CoreDataの変更(undo/redo)を捉える。これにより二重にKVOされるが、Undo/Redoを捕捉できないよりはマシと考える。
        NotificationCenter.default.addObserver(forName: .NSManagedObjectContextObjectsDidChange,
                                               object: document.managedObjectContext,
                                               queue: nil)
        {
            theNotification in
            
            if let dict = theNotification.userInfo
            {
                [
                    dict[NSInsertedObjectsKey] as? Set<NSManagedObject> ,
                    dict[NSUpdatedObjectsKey] as? Set<NSManagedObject>  ,
                    dict[NSUpdatedObjectsKey] as? Set<NSManagedObject>  ,
                    dict[NSDeletedObjectsKey] as? Set<NSManagedObject>
                ]
                .compactMap{$0}                                // unionするためにnilを削除
                .reduce(Set()){$0.union($1)}                   // 1つにunionして重複削除
                .filter(){ $0.entity.name == self.entityName } // 対象の型のManegedObjectだけに絞る
                .forEach()
                {
                    // キー全部(1個しかこないけど)にKVOさせる
                    if let theKeyString = $0.value(forKey: .MINTLDocumentDefaultsKeyNameString) as? String
                    {
                        self.willChangeValue(forKey: theKeyString)
                        self.didChangeValue(forKey: theKeyString)
                    }
                }
            }
        }
    }
    
    // MARK: - dealloc
    
    // MARK: - NSCopying, hash, isEqual:
    
    // MARK: - NSCoding methods
    
    // MARK: - restorableState methods
    
    // MARK: - life cycle methods
    
    // MARK: - action methods
    
    // レコードの存在確認
    @objc func recodeExists(name: String) -> Bool
    {
        return self.findRecode(name: name) != nil
    }

    // レコードの削除。nilを設定するのではなく。削除。
    @objc func deleteRecode(name: String)
    {
        let theFetchRequest = fetchRequest
        
        theFetchRequest.predicate = NSPredicate(format: .MINTLDocumentDefaultsKeyNameString + " == %@", name)

        if let theObject = try! managedObjectContext.fetch(theFetchRequest).last
        {
            self.willChangeValue(forKey: name)

            managedObjectContext.delete(theObject)
            
            self.didChangeValue(forKey: name)
        }
    }

    // KVCでの値の取得時に、構文的に"値"が存在しない場合に呼び出される。
    override func value(forUndefinedKey key: String) -> Any?
    {
        if let theObject = findRecode(name: key)
        {
            // ここで、nilが帰る時があるが、それは正しい動作です。
            return theObject.value(forKey: .MINTLDocumentDefaultsValueNameString)
        }
        else
        {
            // UserDefaultを検索し、あればそれを使う
            if let theUDObject = UserDefaults.standard.object(forKey: key)
            {
                // UserDefaultの値をパススルーで使用していることを記録。UserDefaultが変更された場合は、このkeyでKVOする。
                self.passThroughKeySet.insert(key)
                
                return theUDObject
            }
        }
        
        return nil
    }
    
    // valueにnilが入れると削除にすべきか？。設定項目の削除は別に作成するべき。nilに意味がある場合がある。
    // KVCでの値の設定時に、構文的に"値"が存在しない場合に呼び出される。
    override func setValue(_ value: Any?, forUndefinedKey key: String)
    {
        self.willChangeValue(forKey: key)
     
        var theObject = findRecode(name: key)
            
        if theObject == nil
        {
            theObject = createRecode(name: key)
            
            // UserDefaultの値にパススルーしている事をやめる。UserDeafultが変更されても、このKeyではKVOしないようにする。
            self.passThroughKeySet.remove(key)
        }
        
        if let theValue = value as! NSObject?
        {
            if theValue.isKind(of: NSString.classForCoder())
            {
                theObject!.setValue(theValue, forKey: .MINTLDocumentDefaultsValueNameString)
            }
            else if theValue.isKind(of: NSNumber.classForCoder())
            {
                if let theNumber = value as! NSNumber?
                {
                    theObject!.setValue(theNumber.stringValue, forKey: .MINTLDocumentDefaultsValueNameString)
                }
            }
            else if theValue.isKind(of: NSDate.classForCoder())
            {
                theObject!.setValue(theValue.description, forKey: .MINTLDocumentDefaultsValueNameString)
            }
            else
            {
                theObject!.setValue(theValue, forKey: .MINTLDocumentDefaultsValueNameString)
            }
        }
        else
        {
            theObject!.setValue(nil, forKey: .MINTLDocumentDefaultsValueNameString)
        }
        
        self.didChangeValue(forKey: key)
    }
    
    
    // MARK: - event handling methods
    
    // MARK: - drawing methods
    
    // MARK: - delegate/datasource methods
    
    // MARK: - accessor methods (in pairs)
    
    private weak var document: NSPersistentDocument!
    private var entityName: String!
    private var passThroughKeySet: Set<String> = Set()
    
    private var managedObjectContext: NSManagedObjectContext
    {
        return document!.managedObjectContext!
    }
    
    private var fetchRequest: NSFetchRequest<NSManagedObject>
    {
        return NSFetchRequest<NSManagedObject>(entityName: entityName)
    }
    
    public var allRecodes: Dictionary<String, String?>
    {
        let theAll = try! managedObjectContext.fetch(fetchRequest)

        return Dictionary(uniqueKeysWithValues: theAll.map()
        {
            (i) -> (String, String?) in
            
            let theKey   = i.value(forKey: .MINTLDocumentDefaultsKeyNameString) as! String
            let theValue = i.value(forKey: .MINTLDocumentDefaultsValueNameString) as? String // nilを許容する
            
            return ( theKey, theValue )
        })
    }
    
    // MARK: - Utility methods
    
    @objc private func findRecode(name: String) -> NSManagedObject?
    {
        let theFetchRequest = fetchRequest
        
        theFetchRequest.predicate = NSPredicate(format: .MINTLDocumentDefaultsKeyNameString + " == %@", name)
        
        let theObject = try! managedObjectContext.fetch(theFetchRequest).last
        
        return theObject
    }
    
    @objc private func createRecode(name: String) -> NSManagedObject?
    {
        let theObject = NSEntityDescription.insertNewObject(forEntityName: entityName,
                                                           into: managedObjectContext)
        
        theObject.setValue(name, forKey: .MINTLDocumentDefaultsKeyNameString)
        
        return theObject
    }

}
